# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium
import os
import re
import subprocess
import sys
import tarfile
from pathlib import Path
from shutil import rmtree

import setuptools
from packaging.version import Version
from setuptools import find_namespace_packages, find_packages, setup
from setuptools.command.bdist_wheel import bdist_wheel as _bdist_wheel

if Version(setuptools.__version__) < Version("78.1.1"):
    raise RuntimeError(
        f"\033[31mYour setuptools version (`{setuptools.__version__}`) is too old and incompatible.\n"
        "Please upgrade it with: `pip install -U setuptools`\033[0m"
    )


CUTLASS_VERSION = "3.5.0"
CUTLASS_RELEASE_URL = f"https://github.com/NVIDIA/cutlass/archive/refs/tags/v{CUTLASS_VERSION}.tar.gz"


def _ensure_cutlass_source() -> Path:
    deps_dir = Path("build") / "_deps"
    deps_dir.mkdir(parents=True, exist_ok=True)

    cutlass_root = deps_dir / f"cutlass-v{CUTLASS_VERSION}"
    marker = cutlass_root / ".gptqmodel_complete"
    if marker.exists():
        return cutlass_root.resolve()

    archive_path = deps_dir / f"cutlass-v{CUTLASS_VERSION}.tar.gz"
    if not archive_path.exists():
        _download_with_progress(
            CUTLASS_RELEASE_URL,
            str(archive_path),
            title=f"Downloading CUTLASS v{CUTLASS_VERSION}",
        )

    if cutlass_root.exists():
        rmtree(cutlass_root)

    with tarfile.open(archive_path, "r:gz") as tar:
        extract_kwargs = {"path": deps_dir}
        if sys.version_info >= (3, 12):
            extract_kwargs["filter"] = "data"
        tar.extractall(**extract_kwargs)

    extracted_dir = deps_dir / f"cutlass-{CUTLASS_VERSION}"
    if not extracted_dir.exists():
        raise RuntimeError("Failed to extract CUTLASS archive")

    extracted_dir.rename(cutlass_root)
    marker.touch()
    return cutlass_root.resolve()


# ---------------------------
# Helpers (no torch required)
# ---------------------------

def _read_env(name, default=None):
    v = os.environ.get(name)
    return v if (v is not None and str(v).strip() != "") else default


def _probe_cmd(args):
    try:
        out = subprocess.check_output(args, stderr=subprocess.STDOUT, text=True, timeout=5)
        return out.strip()
    except Exception:
        return None


def _bool_env(name, default=False):
    v = _read_env(name)
    if v is None:
        return default
    return str(v).lower() in ("1", "true", "yes", "y", "on")


def _detect_rocm_version():
    v = _read_env("ROCM_VERSION")
    if v:
        return v
    hip = _probe_cmd(["hipcc", "--version"])
    if hip:
        import re
        m = re.search(r"\b([0-9]+\.[0-9]+)\b", hip)
        if m:
            return m.group(1)
    try:
        p = Path("/opt/rocm/.info/version")
        if p.exists():
            return p.read_text(encoding="utf-8").strip()
    except Exception:
        pass
    return None


def _detect_cuda_arch_list():
    """Return TORCH_CUDA_ARCH_LIST style string for the *installed* GPUs only.
    Priority:
      1) CUDA_ARCH_LIST env override (verbatim)
      2) nvidia-smi compute_cap (actual devices)
    """
    # 1) explicit override
    env_arch = _read_env("CUDA_ARCH_LIST")
    if env_arch:
        return env_arch

    # 2) actual devices present
    smi_out = _probe_cmd(["nvidia-smi", "--query-gpu=compute_cap", "--format=csv,noheader"])
    if smi_out:
        caps = []
        for line in smi_out.splitlines():
            cap = line.strip()
            if not cap:
                continue
            # normalize like '8.0'
            try:
                major, minor = cap.split(".", 1)
                caps.append(f"{int(major)}.{int(minor)}")
            except Exception:
                # some drivers return just '8' -> treat as '8.0'
                if cap.isdigit():
                    caps.append(f"{cap}.0")
        caps = sorted(set(caps), key=lambda x: (int(x.split(".")[0]), int(x.split(".")[1])))
        if caps:
            # PyTorch prefers ';' separators
            return ";".join(caps)

    # 3) conservative default for modern datacenter GPUs (A100 et al.)
    raise Exception("Could not get compute capability from nvidia-smi. Please check nvidia-utils package is installed.")


def _parse_arch_list(s: str):
    # Accept semicolons, commas, and any whitespace as separators.
    # Keep tokens like "8.0", "8.0+PTX" intact (we’ll strip suffixes later).
    return [tok for tok in re.split(r"[;\s,]+", s) if tok.strip()]


def _has_cuda_v8_from_arch_list(arch_list):
    try:
        vals = []
        for a in arch_list:
            # Handle things like "8.0+PTX"
            base = a.split("+", 1)[0]
            vals.append(float(base))
        return any(v >= 8.0 for v in vals)
    except Exception:
        return False


def _detect_cxx11_abi():
    v = _read_env("CXX11_ABI")
    if v in ("0", "1"):
        return int(v)
    return 1


def _torch_version_for_release():
    # No torch import; allow env override
    v = _read_env("TORCH_VERSION")
    if v:
        parts = v.split(".")
        return ".".join(parts[:2])
    else:
        raise Exception("TORCH_VERSION not passed for wheel generation.")
    return None


def _is_rocm_available():
    return _detect_rocm_version() is not None


# If you already have _probe_cmd elsewhere, you can delete this copy.
def _probe_cmd(args, timeout=6):
    try:
        return subprocess.check_output(args, stderr=subprocess.STDOUT, text=True, timeout=timeout)
    except Exception:
        return ""


def _first_token_line(s: str) -> str | None:
    for line in (s or "").splitlines():
        t = line.strip()
        if t:
            return t
    return None


def _detect_torch_version() -> str | None:
    # 1) uv pip show torch
    out = _probe_cmd(["uv", "pip", "show", "torch"])
    if out:
        m = re.search(r"^Version:\s*([^\s]+)\s*$", out, flags=re.MULTILINE)
        if m:
            return m.group(1)

    # 2) pip show torch (both 'pip' and 'python -m pip')
    for cmd in (["pip", "show", "torch"], [sys.executable, "-m", "pip", "show", "torch"]):
        out = _probe_cmd(cmd)
        if out:
            m = re.search(r"^Version:\s*([^\s]+)\s*$", out, flags=re.MULTILINE)
            if m:
                return m.group(1)

    # 3) conda list torch
    out = _probe_cmd(["conda", "list", "torch"])
    if out:
        # Typical line starts with: torch  2.4.1  ...
        for line in out.splitlines():
            if line.strip().startswith("torch"):
                parts = re.split(r"\s+", line.strip())
                if len(parts) >= 2 and re.match(r"^\d+\.\d+(\.\d+)?", parts[1]):
                    return parts[1]

    # 4) Fallback: importlib.metadata (does not import torch package module)
    try:
        import importlib.metadata as im  # py3.8+
        version = im.version("torch")
        if not version:
            raise Exception("torch not found")
    except Exception:
        raise Exception("Unable to detect torch version via uv/pip/conda/importlib. Please install torch >= 2.7.1")


def _detect_torch_accelerator_backend() -> str | None:
    """Return accelerator backend for installed torch build.

    Returns:
      - "cuda" when torch is built with CUDA support
      - "rocm" when torch is built with ROCm support
      - "cpu" when torch is CPU-only
      - None when torch cannot be imported
    """
    try:
        import torch  # type: ignore
    except Exception:
        return None

    version_info = getattr(torch, "version", None)
    if version_info is None:
        return None

    if getattr(version_info, "hip", None):
        return "rocm"
    if getattr(version_info, "cuda", None):
        return "cuda"
    return "cpu"


def _major_minor(v: str) -> str:
    if v:
        parts = v.split(".")
        return ".".join(parts[:2]) if parts else v
    return v


def _version_geq(version: str | None, major: int, minor: int = 0) -> bool:
    if not version:
        return False
    try:
        parts = re.split(r"[._-]", version)
        ver_major = int(parts[0]) if parts else 0
        ver_minor = int(parts[1]) if len(parts) > 1 else 0
        return (ver_major, ver_minor) >= (major, minor)
    except Exception:
        return False


def _nvcc_release_version() -> str | None:
    # Search for nvcc in common locations before giving up.
    candidates: list[str] = []
    nvcc_env = _read_env("NVCC")
    if nvcc_env:
        candidates.append(nvcc_env)

    cuda_home = _read_env("CUDA_HOME")
    cuda_path = _read_env("CUDA_PATH")

    candidates.extend(
        [
            "nvcc",
            str(Path(cuda_home).joinpath("bin", "nvcc")) if cuda_home else None,
            str(Path(cuda_path).joinpath("bin", "nvcc")) if cuda_path else None,
            "/usr/local/cuda/bin/nvcc",
        ]
    )

    seen = set()
    for cmd in candidates:
        if not cmd or cmd in seen:
            continue
        seen.add(cmd)
        out = _probe_cmd([cmd, "--version"])
        if not out:
            continue
        match = re.search(r"release\s+(\d+)\.(\d+)", out)
        if match:
            return f"{match.group(1)}.{match.group(2)}"

    print(
        "NVCC not found (checked PATH, $CUDA_HOME/bin, $CUDA_PATH/bin, /usr/local/cuda/bin). "
        "For Ubuntu, run `sudo update-alternatives --config cuda` to fix path for already installed Cuda."
    )
    return None


def _detect_cuda_version() -> str | None:
    # Priority: env → nvidia-smi → nvcc
    v = os.environ.get("CUDA_VERSION")
    if v and v.strip():
        return v.strip()

    # nvcc --version (parse 'release X.Y')
    return _nvcc_release_version()


def _detect_nvcc_version() -> str | None:
    return _nvcc_release_version()


def get_version_tag() -> str:
    # TODO FIX ME: cpu wheels don't have torch version tags?
    if BUILD_CUDA_EXT != "1":
        return "cpu"

    # TODO FIX ME: rocm wheels don't have torch version tags?
    if ROCM_VERSION:
        return f"rocm{ROCM_VERSION}"

    if not CUDA_VERSION:
        raise Exception("Trying to compile GPTQModel for CUDA/ROCm, but no cuda or rocm version was detected.")

    torch_suffix = f"torch{_major_minor(TORCH_VERSION)}"

    CUDA_VERSION_COMPACT = "".join(CUDA_VERSION.split("."))
    base = f"cu{CUDA_VERSION_COMPACT[:3]}"
    return f"{base}{torch_suffix}"


# ---------------------------
# Env and versioning
# ---------------------------

TORCH_VERSION = _read_env("TORCH_VERSION")
RELEASE_MODE = _read_env("RELEASE_MODE")
CUDA_VERSION = _read_env("CUDA_VERSION")
ROCM_VERSION = _read_env("ROCM_VERSION")
TORCH_CUDA_ARCH_LIST = _read_env("TORCH_CUDA_ARCH_LIST")
NVCC_VERSION = _read_env("NVCC_VERSION")
TORCH_ACCELERATOR_BACKEND = _read_env("TORCH_ACCELERATOR_BACKEND")

# respect user env then detect
if not TORCH_VERSION:
    TORCH_VERSION = _detect_torch_version()
if not CUDA_VERSION:
    CUDA_VERSION = _detect_cuda_version()
if not ROCM_VERSION:
    ROCM_VERSION = _detect_rocm_version()
if not NVCC_VERSION:
    NVCC_VERSION = _detect_nvcc_version()
if not TORCH_ACCELERATOR_BACKEND:
    TORCH_ACCELERATOR_BACKEND = _detect_torch_accelerator_backend()

SKIP_ROCM_VERSION_CHECK = _read_env("SKIP_ROCM_VERSION_CHECK")
FORCE_BUILD = _bool_env("GPTQMODEL_FORCE_BUILD", False)

# BUILD_CUDA_EXT:
# - If user sets explicitly, respect it.
# - Otherwise auto: enable only when torch backend and toolkit match.
BUILD_CUDA_EXT = _read_env("BUILD_CUDA_EXT")
if BUILD_CUDA_EXT is None:
    if TORCH_ACCELERATOR_BACKEND == "cuda":
        BUILD_CUDA_EXT = "1" if CUDA_VERSION else "0"
    elif TORCH_ACCELERATOR_BACKEND == "rocm":
        BUILD_CUDA_EXT = "1" if ROCM_VERSION else "0"
    elif TORCH_ACCELERATOR_BACKEND == "cpu":
        BUILD_CUDA_EXT = "0"
    else:
        BUILD_CUDA_EXT = "1" if (CUDA_VERSION or ROCM_VERSION) else "0"

if ROCM_VERSION and not SKIP_ROCM_VERSION_CHECK:
    try:
        if float(ROCM_VERSION) < 6.2:
            sys.exit(
                "GPTQModel's compatibility with ROCm < 6.2 has not been verified. "
                "Set SKIP_ROCM_VERSION_CHECK=1 to proceed."
            )
    except Exception:
        pass

# Handle CUDA_ARCH_LIST (public) and set TORCH_CUDA_ARCH_LIST for build toolchains
CUDA_ARCH_LIST = _detect_cuda_arch_list() if (BUILD_CUDA_EXT == "1" and not ROCM_VERSION) else None

if not TORCH_CUDA_ARCH_LIST and CUDA_ARCH_LIST:
    archs = _parse_arch_list(CUDA_ARCH_LIST)
    kept = []
    for arch in archs:
        try:
            base = arch.split("+", 1)[0]
            if float(base) >= 6.0:
                kept.append(arch)
            else:
                print(f"we do not support this compute arch: {arch}, skipped.")
        except Exception:
            kept.append(arch)

    # Use semicolons for TORCH_CUDA_ARCH_LIST (PyTorch likes this),
    TORCH_CUDA_ARCH_LIST = ";".join(kept)
    os.environ["TORCH_CUDA_ARCH_LIST"] = TORCH_CUDA_ARCH_LIST

    print(f"CUDA_ARCH_LIST: {CUDA_ARCH_LIST}")
    print(f"TORCH_CUDA_ARCH_LIST: {TORCH_CUDA_ARCH_LIST}")

version_vars = {}
exec("exec(open('gptqmodel/version.py').read()); version=__version__", {}, version_vars)
gptqmodel_version = version_vars["version"]

# -----------------------------
# Prebuilt wheel download config
# -----------------------------
# Default template (GitHub Releases), can be overridden via env.
DEFAULT_WHEEL_URL_TEMPLATE = "https://github.com/ModelCloud/GPTQModel/releases/download/{tag_name}/{wheel_name}"
WHEEL_URL_TEMPLATE = os.environ.get("GPTQMODEL_WHEEL_URL_TEMPLATE")
WHEEL_BASE_URL = os.environ.get("GPTQMODEL_WHEEL_BASE_URL")
WHEEL_TAG = os.environ.get("GPTQMODEL_WHEEL_TAG")  # Optional override of release tag


def _resolve_wheel_url(tag_name: str, wheel_name: str) -> str:
    """
    Build the final wheel URL based on:
      1) GPTQMODEL_WHEEL_URL_TEMPLATE (highest priority)
      2) GPTQMODEL_WHEEL_BASE_URL (append /{wheel_name})
      3) DEFAULT_WHEEL_URL_TEMPLATE (GitHub Releases)
    """
    # Highest priority: explicit template
    if WHEEL_URL_TEMPLATE:
        tmpl = WHEEL_URL_TEMPLATE
        # If {wheel_name} or {tag_name} not present, treat as base and append name.
        if ("{wheel_name}" in tmpl) or ("{tag_name}" in tmpl):
            return tmpl.format(tag_name=tag_name, wheel_name=wheel_name)
        # Otherwise, join as base
        if tmpl.endswith("/"):
            return tmpl + wheel_name
        return tmpl + "/" + wheel_name

    # Next priority: base URL
    if WHEEL_BASE_URL:
        base = WHEEL_BASE_URL
        if base.endswith("/"):
            return base + wheel_name
        return base + "/" + wheel_name

    # Fallback: default GitHub template
    return DEFAULT_WHEEL_URL_TEMPLATE.format(tag_name=tag_name, wheel_name=wheel_name)


def _download_with_progress(url: str, dest_path: str, title: str = "Downloading") -> None:
    """Download url to dest_path with simple stdout progress updates."""
    import time
    import urllib.request as req

    start_time = time.time()
    last_draw_time = 0.0
    last_print_percent = -1

    def _format_bytes(num_bytes: float) -> str:
        units = ["B", "KiB", "MiB", "GiB", "TiB"]
        value = float(max(num_bytes, 0.0))
        for unit in units:
            if value < 1024.0 or unit == units[-1]:
                return f"{value:0.1f}{unit}" if unit != "B" else f"{int(value)}B"
            value /= 1024.0
        return f"{value:0.1f}TiB"

    def _reporthook(block_num: int, block_size: int, total_size: int) -> None:
        nonlocal last_draw_time, last_print_percent
        now = time.time()
        downloaded = block_num * block_size
        speed = downloaded / max(now - start_time, 1e-6)

        if total_size and total_size > 0:
            percent = min(int(downloaded * 100 / total_size), 100)
            if percent == last_print_percent and percent != 100:
                return
            subtitle = (
                f"{percent:3d}% ({_format_bytes(downloaded)}/{_format_bytes(total_size)}) "
                f"{_format_bytes(speed)}/s"
            )
            print(f"{title} {subtitle}", flush=True)
            last_print_percent = percent
            last_draw_time = now
        else:
            if (now - last_draw_time) < 1.0:
                return
            subtitle = f"{_format_bytes(downloaded)} {_format_bytes(speed)}/s"
            print(f"{title} {subtitle}", flush=True)
            last_draw_time = now

    req.urlretrieve(url, dest_path, reporthook=_reporthook)


# Decide HAS_CUDA_V8 / HAS_CUDA_V9 without torch
HAS_CUDA_V8 = False
HAS_CUDA_V9 = False
if CUDA_ARCH_LIST:
    arch_list = _parse_arch_list(CUDA_ARCH_LIST)
    try:
        caps = [float(tok.split("+", 1)[0]) for tok in arch_list]
    except Exception:
        caps = []
    if not ROCM_VERSION:
        HAS_CUDA_V8 = any(cap >= 8.0 for cap in caps)
        HAS_CUDA_V9 = any(cap >= 9.0 for cap in caps)
else:
    smi = _probe_cmd(["nvidia-smi", "--query-gpu=compute_cap", "--format=csv,noheader"])
    if smi:
        try:
            caps = [float(x.strip()) for x in smi.splitlines() if x.strip()]
            HAS_CUDA_V8 = any(cap >= 8.0 for cap in caps)
            HAS_CUDA_V9 = any(cap >= 9.0 for cap in caps)
        except Exception:
            HAS_CUDA_V8 = False
            HAS_CUDA_V9 = False

if RELEASE_MODE == "1":
    gptqmodel_version = f"{gptqmodel_version}+{get_version_tag()}"

include_dirs = ["gptqmodel_ext"]

extensions = []
additional_setup_kwargs = {}


# ---------------------------
# Build CUDA/ROCm extensions (only when enabled)
# ---------------------------
# -----------------------------
# Per-extension build toggles
# -----------------------------
def _env_enabled(val: str) -> bool:
    if val is None:
        return True
    return str(val).strip().lower() not in ("0", "false", "off", "no")


def _env_enabled_any(names, default="1") -> bool:
    for n in names:
        if n in os.environ:
            return _env_enabled(os.environ.get(n))
    return _env_enabled(default)


BUILD_MARLIN = _env_enabled_any(os.environ.get("GPTQMODEL_BUILD_MARLIN", "1"))
BUILD_MACHETE = _env_enabled(os.environ.get("GPTQMODEL_BUILD_MACHETE", "0"))
BUILD_EXLLAMA_V2 = _env_enabled(os.environ.get("GPTQMODEL_BUILD_EXLLAMA_V2", "1"))
BUILD_QQQ = _env_enabled(os.environ.get("GPTQMODEL_BUILD_QQQ", "1"))
BUILD_AWQ = _env_enabled(os.environ.get("GPTQMODEL_BUILD_AWQ", "1"))

# Optional kernels and not build by default. Enable compile with env flags
BUILD_EORA = _env_enabled(os.environ.get("GPTQMODEL_BUILD_EORA", "0"))

if BUILD_CUDA_EXT == "1":
    # Import torch's cpp_extension only if we're truly building GPU extensions
    try:

        from torch.utils import cpp_extension as cpp_ext  # type: ignore
    except Exception:
        if FORCE_BUILD:
            sys.exit(
                "FORCE_BUILD is set but PyTorch C++ extension headers are unavailable. "
                "Install torch build deps first (see https://pytorch.org/) or unset GPTQMODEL_FORCE_BUILD."
            )
        # If we can't import cpp_extension, fall back to prebuilt wheel path
        cpp_ext = None

    if cpp_ext is not None:
        # Limit compile parallelism to avoid overwhelming nvcc/cicc invocations.
        # Respect pre-set MAX_JOBS, otherwise fall back to CPU count minus two (min 1).
        cpu_count = os.cpu_count() or 1
        default_max_jobs = max(1, cpu_count - 2)
        max_jobs_raw = os.environ.get("MAX_JOBS")
        if max_jobs_raw is None or max_jobs_raw.strip() == "":
            effective_max_jobs = default_max_jobs
            print(f"MAX_JOBS not set; defaulting to {effective_max_jobs} concurrent CUDA compilations.")
        else:
            try:
                parsed_jobs = int(max_jobs_raw)
            except ValueError:
                effective_max_jobs = default_max_jobs
                print(f"Ignoring invalid MAX_JOBS={max_jobs_raw!r}; using {effective_max_jobs}.")
            else:
                if parsed_jobs <= 0:
                    effective_max_jobs = default_max_jobs
                    print(f"MAX_JOBS={parsed_jobs} is non-positive; using {effective_max_jobs}.")
                else:
                    effective_max_jobs = parsed_jobs

        os.environ["MAX_JOBS"] = str(effective_max_jobs)
        os.environ["NINJA_NUM_JOBS"] = str(effective_max_jobs)
        print(f"Using MAX_JOBS={effective_max_jobs} to cap concurrent CUDA compilations.")

        nvcc_threads = 2
        os.environ["NVCC_THREADS"] = str(nvcc_threads)
        print(f"Using NVCC_THREADS={nvcc_threads} for per-invocation NVCC concurrency.")

        # Optional conda CUDA runtime headers
        # conda_cuda_include_dir = os.path.join(get_python_lib(), "nvidia/cuda_runtime/include")
        # if os.path.isdir(conda_cuda_include_dir):
        #     include_dirs.append(conda_cuda_include_dir)
        #     print(f"appending conda cuda include dir {conda_cuda_include_dir}")

        extra_link_args = []
        extra_compile_args = {
            "cxx": ["-O3", "-std=c++17", "-DENABLE_BF16"],
            "nvcc": [
                "-O3",
                "-std=c++17",
                "-DENABLE_BF16",
                "-U__CUDA_NO_HALF_OPERATORS__",
                "-U__CUDA_NO_HALF_CONVERSIONS__",
                "-U__CUDA_NO_BFLOAT16_OPERATORS__",
                "-U__CUDA_NO_BFLOAT16_CONVERSIONS__",
                "-U__CUDA_NO_BFLOAT162_OPERATORS__",
                "-U__CUDA_NO_BFLOAT162_CONVERSIONS__",
            ],
        }

        cutlass_include_paths = []
        if BUILD_MACHETE:
            cutlass_root = _ensure_cutlass_source()
            cutlass_include_paths = [
                Path("gptqmodel_ext/cutlass_extensions").resolve(),
                cutlass_root / "include",
                cutlass_root / "examples/common/include",
                cutlass_root / "tools/library/include",
            ]
            if "GPTQMODEL_CUTLASS_DIR" not in os.environ:
                os.environ["GPTQMODEL_CUTLASS_DIR"] = str(cutlass_root)
            cutlass_include_flags = [f"-I{path}" for path in cutlass_include_paths]
            extra_compile_args["cxx"] += cutlass_include_flags
            extra_compile_args["nvcc"] += cutlass_include_flags

        # Windows/OpenMP note: adjust flags as needed for MSVC if you add native Windows wheels
        if sys.platform == "win32":
            extra_compile_args["cxx"] = ["/O2", "/std:c++17", "/openmp", "/DNDEBUG", "/DENABLE_BF16"]

        CXX11_ABI = _detect_cxx11_abi()
        extra_compile_args["cxx"] += [f"-D_GLIBCXX_USE_CXX11_ABI={CXX11_ABI}"]
        extra_compile_args["nvcc"] += [f"-D_GLIBCXX_USE_CXX11_ABI={CXX11_ABI}"]

        if not ROCM_VERSION:
            # if _version_geq(NVCC_VERSION, 13, 0):
            #     extra_compile_args["nvcc"].append("--device-entity-has-hidden-visibility=false")
            nvcc_extra_flags = [
                "--threads", str(nvcc_threads),  # NVCC parallelism
                "--optimize=3",  # alias for -O3
                # "-rdc=true",  # enable relocatable device code, required for future cuda > 13.x <-- TODO FIX ME broken loading
                # "-dlto",      # compile and link <-- TODO FIX ME
                # Print register/shared-memory usage per kernel (debug aid, no perf effect)
                # Ensure PTXAS uses maximum optimization
                # Cache global loads in both L1 and L2 (better for memory-bound kernels)
                "-Xptxas", "-v,-O3,-dlcm=ca",
                "-lineinfo",  # keep source line info for profiling
                # "--resource-usage",  # show per-kernel register/SMEM usage
                "-Xfatbin", "-compress-all",  # compress fatbin
                # "--expt-relaxed-constexpr",  # relaxed constexpr rules <-- not used
                # "--expt-extended-lambda",  # allow device lambdas <-- not used
                "-diag-suppress=179,39,177",  # silence some template warnings
            ]
            if _version_geq(NVCC_VERSION, 12, 8):
                # Allow instantiations of __global__ templates to live in different TUs; only supported in newer NVCC.
                nvcc_extra_flags.insert(0, "-static-global-template-stub=false")
            extra_compile_args["nvcc"] += nvcc_extra_flags
        else:
            # hipify CUDA-like flags
            def _hipify_compile_flags(flags):
                modified_flags = []
                for flag in flags:
                    if flag.startswith("-") and "CUDA" in flag and not flag.startswith("-I"):
                        parts = flag.split("=", 1)
                        if len(parts) == 2:
                            flag_part, value_part = parts
                            modified_flag_part = flag_part.replace("CUDA", "HIP", 1)
                            modified_flags.append(f"{modified_flag_part}={value_part}")
                        else:
                            modified_flags.append(flag.replace("CUDA", "HIP", 1))
                    else:
                        modified_flags.append(flag)
                return modified_flags


            extra_compile_args["nvcc"] = _hipify_compile_flags(extra_compile_args["nvcc"])

        # Extensions (gate marlin/qqq/eora/exllamav2 on CUDA sm_80+ and non-ROCm)
        if sys.platform != "win32":
            if not ROCM_VERSION and HAS_CUDA_V8:
                if BUILD_MARLIN:
                    marlin_kernel_dir = Path("gptqmodel_ext/marlin")
                    marlin_kernel_files = sorted(marlin_kernel_dir.glob("kernel_*.cu"))

                    if not marlin_kernel_files:
                        generator_script = marlin_kernel_dir / "generate_kernels.py"
                        if generator_script.exists():
                            print("Regenerating marlin template instantiations for parallel compilation...")
                            subprocess.check_call([sys.executable, str(generator_script)])
                            marlin_kernel_files = sorted(marlin_kernel_dir.glob("kernel_*.cu"))

                    if not marlin_kernel_files:
                        raise RuntimeError(
                            "No generated marlin kernel templates detected. Run generate_kernels.py before building."
                        )

                    marlin_template_kernel_srcs = [str(path) for path in marlin_kernel_files]
                    extensions += [
                        cpp_ext.CUDAExtension(
                            "gptqmodel_marlin_kernels",
                            [
                                "gptqmodel_ext/marlin/marlin_cuda.cpp",
                                "gptqmodel_ext/marlin/gptq_marlin.cu",
                                "gptqmodel_ext/marlin/gptq_marlin_repack.cu",
                                "gptqmodel_ext/marlin/awq_marlin_repack.cu",
                            ] + marlin_template_kernel_srcs,
                            extra_link_args=extra_link_args,
                            extra_compile_args=extra_compile_args,
                        )
                    ]

                if BUILD_MACHETE and HAS_CUDA_V9 and _version_geq(NVCC_VERSION, 12, 0):
                    try:
                        result = subprocess.run(
                            [sys.executable, "gptqmodel_ext/machete/generate.py"],
                            check=True,
                            text=True,
                            capture_output=True
                        )
                    except subprocess.CalledProcessError as e:
                        raise RuntimeError(
                            f"Error generating machete kernel templates:\n"
                            f"Return code: {e.returncode}\n"
                            f"Stderr: {e.stderr}\n"
                            f"Stdout: {e.stdout}"
                        )
                    machete_dir = Path("gptqmodel_ext/machete")
                    machete_generated_dir = machete_dir / "generated"

                    machete_sources = [str(machete_dir / "machete_pytorch.cu")]
                    machete_generated_sources = sorted(machete_generated_dir.glob("*.cu"))

                    if not machete_generated_sources:
                        raise RuntimeError(
                            "No generated machete kernel templates detected. Run gptqmodel_ext/machete/generate.py"
                            " with CUTLASS checkout before building."
                        )

                    machete_sources += [str(path) for path in machete_generated_sources]

                    machete_include_dirs = [str(Path("gptqmodel_ext").resolve())] + [str(path) for path in cutlass_include_paths]

                    extensions += [
                        cpp_ext.CUDAExtension(
                            "gptqmodel_machete_kernels",
                            machete_sources,
                            extra_link_args=extra_link_args,
                            extra_compile_args=extra_compile_args,
                            include_dirs=machete_include_dirs,
                        )
                    ]

                if BUILD_QQQ:
                    extensions += [
                        cpp_ext.CUDAExtension(
                            "gptqmodel_qqq_kernels",
                            [
                                "gptqmodel_ext/qqq/qqq.cpp",
                                "gptqmodel_ext/qqq/qqq_gemm.cu",
                            ],
                            extra_link_args=extra_link_args,
                            extra_compile_args=extra_compile_args,
                        )
                    ]

                if BUILD_EORA:
                    extensions += [
                        cpp_ext.CUDAExtension(
                            "gptqmodel_exllama_eora",
                            [
                                "gptqmodel_ext/exllama_eora/eora/q_gemm.cu",
                                "gptqmodel_ext/exllama_eora/eora/pybind.cu",
                            ],
                            extra_link_args=extra_link_args,
                            extra_compile_args=extra_compile_args,
                        )
                    ]
                if BUILD_EXLLAMA_V2:
                    extensions += [
                        cpp_ext.CUDAExtension(
                            "gptqmodel_exllamav2_kernels",
                            [
                                "gptqmodel_ext/exllamav2/ext_gptq.cpp",
                                "gptqmodel_ext/exllamav2/cuda/q_matrix.cu",
                                "gptqmodel_ext/exllamav2/cuda/q_gemm.cu",
                            ],
                            extra_link_args=extra_link_args,
                            extra_compile_args=extra_compile_args,
                        )
                    ]

            if BUILD_AWQ:
                if ROCM_VERSION:
                    print("Skipping AWQ kernels on ROCm: inline PTX is CUDA-only.")
                else:
                    extensions += [
                        # contain un-hipifiable inline PTX
                        cpp_ext.CUDAExtension(
                            "gptqmodel_awq_kernels",
                            [
                                "gptqmodel_ext/awq/pybind_awq.cpp",
                                "gptqmodel_ext/awq/quantization/gemm_cuda_gen.cu",
                                "gptqmodel_ext/awq/quantization/gemv_cuda.cu",
                            ],
                            extra_link_args=extra_link_args,
                            extra_compile_args=extra_compile_args,
                        ),
                        # TODO only compatible with ampere?
                        # arch_flags = get_compute_capabilities({80, 86, 89, 90})
                        # extra_compile_args_v2 = get_extra_compile_args(arch_flags, generator_flags)
                        cpp_ext.CUDAExtension(
                            "gptqmodel_awq_v2_kernels",
                            [
                                "gptqmodel_ext/awq/pybind_awq_v2.cpp",
                                "gptqmodel_ext/awq/quantization_new/gemv/gemv_cuda.cu",
                                "gptqmodel_ext/awq/quantization_new/gemm/gemm_cuda.cu",
                            ],
                            extra_link_args=extra_link_args,
                            extra_compile_args=extra_compile_args,
                        ),
                        cpp_ext.CUDAExtension(
                            "gptqmodel_exllamav2_awq_kernels",
                            [
                                "gptqmodel_ext/exllamav2/ext_awq.cpp",
                                "gptqmodel_ext/exllamav2/cuda/q_matrix_awq.cu",
                                "gptqmodel_ext/exllamav2/cuda/q_gemm_awq.cu",
                            ],
                            extra_link_args=extra_link_args,
                            extra_compile_args=extra_compile_args,
                        )
                    ]

        # Ensure machete kernels are compiled before other extensions
        machete_exts = [ext for ext in extensions if getattr(ext, "name", "") == "gptqmodel_machete_kernels"]
        if machete_exts:
            other_exts = [ext for ext in extensions if getattr(ext, "name", "") != "gptqmodel_machete_kernels"]
            extensions[:] = machete_exts + other_exts

        additional_setup_kwargs = {
            "ext_modules": extensions,
            "cmdclass": {"build_ext": cpp_ext.BuildExtension},
        }

        # additional_setup_kwargs = {
        #     "ext_modules": extensions,
        #     # "include_package_data": True,
        #     # "package_data": {"": ["build/lib/*.so"]},
        #     "cmdclass": {"build_ext": cpp_ext.BuildExtension.with_options(
        #         use_ninja=True,
        #         no_python_abi_suffix=True,
        #         build_temp="build/temp",
        #         # build_lib="build/lib", TODO FIX ME why package_data doesn't work..
        #         clean_first=False  # keep intermediates for reuse
        #     )},
        # }


# ---------------------------
# Cached wheel fetcher
# ---------------------------

class CachedWheelsCommand(_bdist_wheel):
    def run(self):
        # No implicit torch checks; allow explicit override via env
        xpu_avail = _bool_env("XPU_AVAILABLE", False)
        if FORCE_BUILD or xpu_avail:
            return super().run()

        python_version = f"cp{sys.version_info.major}{sys.version_info.minor}{sys.abiflags}"

        wheel_filename = f"gptqmodel-{gptqmodel_version}+{get_version_tag()}-{python_version}-{python_version}-linux_x86_64.whl"

        # Allow tag override via env; default to "v{gptqmodel_version}"
        tag_name = WHEEL_TAG if WHEEL_TAG else f"v{gptqmodel_version}"
        wheel_url = _resolve_wheel_url(tag_name=tag_name, wheel_name=wheel_filename)

        print(f"Resolved wheel URL: {wheel_url}\nwheel name={wheel_filename}")

        try:
            if not os.path.exists(self.dist_dir):
                os.makedirs(self.dist_dir)

            wheel_path = os.path.join(self.dist_dir, wheel_filename)

            _download_with_progress(wheel_url, wheel_path, title="Downloading wheel")
            print("Raw wheel path", wheel_filename)
        except BaseException:
            env_info = [f"python={python_version}", f"torch={TORCH_VERSION or 'unknown'}"]
            if CUDA_VERSION:
                env_info.append(f"cuda={CUDA_VERSION}")
            if ROCM_VERSION:
                env_info.append(f"rocm={ROCM_VERSION}")

            print(
                "Unable to match and download a precompiled wheel; entering slow manual build mode. "
                f"Wheel match params: {', '.join(env_info)}. "
                f"Fallback source build triggered for {wheel_url}"
            )
            super().run()


# ---------------------------
# setup()
# ---------------------------
print(f"CUDA {CUDA_ARCH_LIST}")
print(f"HAS_CUDA_V8 {HAS_CUDA_V8}")
print(f"HAS_CUDA_V9 {HAS_CUDA_V9}")
print(f"SETUP_KWARGS {additional_setup_kwargs}")
print(f"gptqmodel_version={gptqmodel_version}")

_namespace_packages = find_namespace_packages(include=["gptqmodel_ext.*"])
_packages = find_packages()
for _pkg in _namespace_packages:
    if _pkg not in _packages:
        _packages.append(_pkg)

setup(
    version=gptqmodel_version,
    packages=_packages,
    include_package_data=True,
    include_dirs=include_dirs,
    cmdclass=(
        {"bdist_wheel": CachedWheelsCommand, "build_ext": additional_setup_kwargs.get("cmdclass", {}).get("build_ext")}
        if (BUILD_CUDA_EXT == "1" and additional_setup_kwargs)
        else {"bdist_wheel": CachedWheelsCommand}
    ),
    ext_modules=additional_setup_kwargs.get("ext_modules", []),
)



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

from ..base import BaseQModel
from ..moe_lifecycle import GateUpDownMoELifecycleHooks


class HymbaQModel(BaseQModel):
    supports_desc_act = [False]
    require_trust_remote_code = True
    require_monkeypatch = True
    require_pkgs = ["tiktoken>=0.7.0",
                            "sentencepiece>=0.2.0",
                            "protobuf>=5.28.3",
                            "ninja>=1.11.1.1",
                            "einops>=0.8.0",
                            "mamba_ssm>=2.2.2",
                            "causal_conv1d>=1.4.0",
                            "attn_gym>=0.0.3.dev5"]

    pre_lm_head_norm_module = "model.final_layernorm"

    # MoE lifecycle hooks for gate_proj/up_proj/down_proj pattern
    moe_lifecycle_hooks = GateUpDownMoELifecycleHooks()

    module_tree = [
        "model",
        "layers",
        "#",
        {
            "input_layernorm": ("input_layernorm:!",),
            "mamba": ("in_proj:0", "out_proj:1"),
            "post_attention_layernorm": ("post_attention_layernorm:!",),
            "moe:moe": {
                "experts": {
                    "0": ("gate_proj:0", "up_proj:0", "down_proj:1"),
                }
            }
        }
    ]

    def monkey_patch(self):
        if hasattr(self.config, 'conv_dim'):
            new_conv_dim = {}
            try:
                for k, v in self.config.conv_dim.items():
                    if isinstance(k, str):
                        new_conv_dim[int(k)] = v
                self.config.conv_dim = new_conv_dim
            except ValueError as e:
                if "invalid literal" in str(e):
                    raise ValueError("The key of HymbaConfig.conv_dim should be a string of numbers.")



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

import math

import torch

from ...adapter.adapter import Adapter
from ...quantization import FORMAT, METHOD
from ...quantization.awq.utils.packing_utils import (
    dequantize_gemm,
    reverse_awq_order,
    unpack_awq,
)
from ...utils.backend import BACKEND
from ...utils.logger import setup_logger
from .gemm_hf_kernel import HFKernelLinear


log = setup_logger()


class HFKernelAwqLinear(HFKernelLinear):
    """AWQ variant of HFKernelLinear — uses kernels-community gemm_int4 with AWQ weights."""

    QUANT_TYPE = "hf_kernel_awq"

    SUPPORTS_BACKENDS = [BACKEND.HF_KERNEL_AWQ]
    SUPPORTS_METHODS = [METHOD.AWQ]
    SUPPORTS_FORMATS = {FORMAT.GEMM: 110}

    # inherit from HFKernelLinear
    SUPPORTS_BITS = HFKernelLinear.SUPPORTS_BITS
    SUPPORTS_GROUP_SIZE = HFKernelLinear.SUPPORTS_GROUP_SIZE
    SUPPORTS_DESC_ACT = HFKernelLinear.SUPPORTS_DESC_ACT
    SUPPORTS_SYM = HFKernelLinear.SUPPORTS_SYM
    SUPPORTS_SHARDS = HFKernelLinear.SUPPORTS_SHARDS
    SUPPORTS_TRAINING = HFKernelLinear.SUPPORTS_TRAINING
    SUPPORTS_AUTO_PADDING = HFKernelLinear.SUPPORTS_AUTO_PADDING
    SUPPORTS_IN_FEATURES_DIVISIBLE_BY = HFKernelLinear.SUPPORTS_IN_FEATURES_DIVISIBLE_BY
    SUPPORTS_OUT_FEATURES_DIVISIBLE_BY = HFKernelLinear.SUPPORTS_OUT_FEATURES_DIVISIBLE_BY
    SUPPORTS_DEVICES = HFKernelLinear.SUPPORTS_DEVICES
    SUPPORTS_PLATFORM = HFKernelLinear.SUPPORTS_PLATFORM
    SUPPORTS_PACK_DTYPES = HFKernelLinear.SUPPORTS_PACK_DTYPES
    SUPPORTS_ADAPTERS = HFKernelLinear.SUPPORTS_ADAPTERS
    REQUIRES_FORMAT_V2 = HFKernelLinear.REQUIRES_FORMAT_V2

    SUPPORTS_DTYPES = [torch.float16, torch.bfloat16]

    def __init__(
        self,
        bits: int,
        group_size: int,
        sym: bool,
        desc_act: bool,
        in_features: int,
        out_features: int,
        bias: bool = False,
        pack_dtype: torch.dtype = torch.int32,
        adapter: Adapter = None,
        register_buffers: bool = True,
        **kwargs,
    ):
        kwargs.setdefault("backend", BACKEND.HF_KERNEL_AWQ)
        super().__init__(
            bits=bits,
            group_size=group_size,
            sym=sym,
            desc_act=desc_act,
            in_features=in_features,
            out_features=out_features,
            bias=bias,
            pack_dtype=pack_dtype,
            adapter=adapter,
            # Skip base buffer init, we need to manually init buffers for awq
            register_buffers=False,
            enable_wf_unsqueeze=kwargs.pop("enable_wf_unsqueeze", False),
            **kwargs,
        )

        # Create awq buffers
        if register_buffers:
            pack_cols = max(1, self.out_features // self.pack_factor)
            qweight_shape = (self.in_features, pack_cols)
            group_size = max(int(self.group_size), 1)
            group_rows = max(1, math.ceil(self.in_features / group_size))

            self.register_buffer(
                "qweight",
                torch.zeros(qweight_shape, dtype=self.pack_dtype),
            )

            self.register_buffer(
                "qzeros",
                torch.zeros((group_rows, pack_cols), dtype=self.pack_dtype),
            )

            self.register_buffer(
                "scales",
                torch.zeros((group_rows, self.out_features), dtype=torch.float16),
            )

            if bias:
                self.register_buffer("bias", torch.zeros(self.out_features, dtype=torch.float16))
            else:
                self.bias = None

    def post_init(self):
        super().post_init()

    def transform_cpu(self):
        # Unpack AWQ weights directly to integer form
        iweight, izeros = unpack_awq(self.qweight, self.qzeros, self.bits)
        iweight, izeros = reverse_awq_order(iweight, izeros, self.bits)
        max_val = (1 << self.bits) - 1
        iweight = torch.bitwise_and(iweight, max_val).to(torch.uint8)
        izeros = torch.bitwise_and(izeros, max_val).to(torch.uint8)

        self.scales = self.scales.to(torch.bfloat16).contiguous()

        # AWQ has no g_idx — weights are already in natural order, just transpose
        # iweight: (in_features, out_features) -> (out_features, in_features)
        self.qweight = iweight.t().contiguous()
        self.qzeros = izeros.contiguous()

    def transform(self, device):
        if device == "cpu":
            self.transform_cpu()
            self.convert_weight_packed_zp()
        else:
            raise NotImplementedError(
                "HFKernelAwqLinear only supports fused transforms on CPU devices."
            )

    def awq_weight_dequantize(self, device, dtype):
        return dequantize_gemm(
            qweight=self.qweight,
            qzeros=self.qzeros,
            scales=self.scales,
            bits=self.bits,
            group_size=self.group_size,
        ).to(device=device, dtype=dtype)

    @torch.no_grad()
    def _fused_op_forward(self, x):
        # AWQ has no g_idx reordering — skip ret_idx
        if x.device.type == "cpu":
            out = self.gemm_int4_forward_kernel(x, self.qweight, self.qzeros, self.scales, self.group_size)
        else:
            raise NotImplementedError
        return out

    def forward(self, x: torch.Tensor):
        out_shape = x.shape[:-1] + (self.out_features,)
        x = x.reshape(-1, x.shape[-1])
        if not self.training and not x.requires_grad and self.linear_mode is None and self.gemm_int4_forward_kernel is not None and x.device.type == "cpu":
            self.transform(x.device.type)
            self.linear_mode = "inference"
        elif self.linear_mode is None:
            self.linear_mode = "train"

        if self.linear_mode == "inference":
            out = self._fused_op_forward(x).reshape(out_shape)
        else:
            weight = self.awq_weight_dequantize(device=x.device, dtype=x.dtype)
            out = torch.matmul(x, weight).reshape(out_shape)

        if self.bias is not None:
            out.add_(self.bias)
        if self.adapter:
            out = self.adapter.apply(x=x, out=out)

        return out


__all__ = ["HFKernelAwqLinear"]



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

import platform
import sys

from packaging.version import Version

from gptqmodel.utils.logger import setup_logger


log = setup_logger()

# Check if GIL (global interpreter lock) is controllable in this Python build.
# Starting from python 3.13 it is possible to disable GIL
def has_gil_control():
    return hasattr(sys, '_is_gil_enabled')

# Check if GIL (global interpreter lock) is enabled.
# Starting from python 3.13 it is possible to disable GIL
def has_gil_disabled():
    return has_gil_control() and not sys._is_gil_enabled()

# Check For Python >= 3.13.3
def gte_python_3_13_3():
    return Version(platform.python_version()) >= Version("3.13.3")

# Check For Python >= 3.14
def gte_python_3_14():
    return Version(platform.python_version()) >= Version("3.14")

# torch compile requires GIL=1 or python 3.13.3t with GIL=0
def log_gil_requirements_for(feature: str):
    log.warn.once(f"Feature `{feature}` requires Python < 3.14 and Python GIL enabled and Python >= 3.13.3T (T for Threading-Free edition of Python) plus Torch 2.8. Feature is currently skipped/disabled.")



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium
from __future__ import annotations

from enum import Enum

import torch
import torch.nn as nn
import transformers
from torch import device

from ..utils import BACKEND
from ..utils.rocm import IS_ROCM
from ..utils.torch import HAS_CUDA, HAS_MPS, HAS_XPU


CPU = device("cpu")
META = device("meta")
CUDA = device("cuda")
CUDA_0 = device("cuda:0")
XPU = device("xpu")
XPU_0 = device("xpu:0")
MPS = device("mps")
ROCM = device("cuda:0")  # rocm maps to fake cuda

SUPPORTS_MODULE_TYPES = [nn.Linear, nn.Conv1d, nn.Conv2d, transformers.Conv1D]

DEFAULT_MAX_SHARD_SIZE = "4GB"

class DEVICE(str, Enum):
    ALL = "all"  # All device
    CPU = "cpu"  # All CPU: Optimized for IPEX is CPU has AVX, AVX512, AMX, or XMX instructions
    CUDA = "cuda"  # Nvidia GPU: Optimized for Ampere+
    XPU = "xpu"  # Intel GPU: Datacenter Max + Arc
    MPS = "mps"  # MacOS GPU: Apple Silicon/Metal)
    ROCM = "rocm"  # AMD GPU: ROCm maps to fake cuda

    @classmethod
    # conversion method called for init when string is passed, i.e. Device("CUDA")
    def _missing_(cls, value):
        if IS_ROCM and f"{value}".lower() == "rocm":
            return cls.ROCM
        return super()._missing_(value)

    @property
    def type(self) -> str:
        """Return the backend type compatible with torch.device semantics."""
        if self == DEVICE.ROCM:
            return "cuda"
        return self.value

    @property
    def index(self) -> int | None:
        """Default index used when materialising a torch.device from this enum."""
        if self in (DEVICE.CUDA, DEVICE.ROCM, DEVICE.XPU):
            return 0
        return None

    def to_torch_device(self) -> torch.device:
        """Convert the enum to a concrete torch.device, defaulting to index 0."""
        idx = self.index
        return torch.device(self.type if idx is None else f"{self.type}:{idx}")

    def to_device_map(self):
        return {"": DEVICE.CUDA if self == DEVICE.ROCM else self}


class PLATFORM(str, Enum):
    ALL = "all"  # All platform
    LINUX = "linux"  # linux
    WIN32 = "win32"  # windows
    DARWIN = "darwin"  # macos


def validate_cuda_support(raise_exception: bool = False):
    got_cuda = HAS_CUDA
    if got_cuda:
        at_least_one_cuda_v6 = any(
            torch.cuda.get_device_capability(i)[0] >= 6 for i in range(torch.cuda.device_count()))

        if not at_least_one_cuda_v6:
            if raise_exception:
                raise EnvironmentError(
                    "GPTQModel cuda requires Pascal or later gpu with compute capability >= `6.0`.")
            else:
                got_cuda = False

    return got_cuda


def normalize_device(type_value: str | DEVICE | int | torch.device) -> DEVICE:
    if isinstance(type_value, int):
        if HAS_CUDA:
            return DEVICE.CUDA
        elif HAS_XPU:
            return DEVICE.XPU
        elif HAS_MPS:
            return DEVICE.MPS
        else:
            return DEVICE.CPU

    if isinstance(type_value, torch.device):
        type_value = type_value.type

    # remove device index
    split_results = [s.strip() for s in type_value.split(":") if s]
    if len(split_results) > 1:
        type_value = split_results[0]

    if isinstance(type_value, DEVICE):
        return type_value

    if not isinstance(type_value, str):
        raise ValueError(f"Invalid device type_value type: {type(type_value)}")

    return DEVICE(type_value.lower())


def get_best_device(backend: BACKEND = BACKEND.AUTO) -> torch.device:
    if HAS_CUDA:
        return CUDA_0
    elif HAS_XPU:
        return XPU_0
    elif HAS_MPS:
        return MPS
    else:
        return CPU

EXPERT_INDEX_PLACEHOLDER = "{expert_index}"



import argparse
from typing import Iterable

from gptqmodel.cli.env import _handle_env_command


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="gptqmodel", description="Utilities for GPTQModel")
    subparsers = parser.add_subparsers(dest="command")

    env_parser = subparsers.add_parser("env", help="Inspect the local GPTQModel runtime environment")
    env_parser.set_defaults(func=_handle_env_command)

    return parser


def main(argv: Iterable[str] | None = None) -> None:
    parser = build_parser()
    args = parser.parse_args(list(argv) if argv is not None else None)

    if hasattr(args, "func"):
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":  # pragma: no cover - manual execution entrypoint
    main()



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

from ..base import BaseQModel


class Ernie4_5_MoeQModel(BaseQModel):
    require_trust_remote_code = True
    support_batch_quantize = False
    require_monkeypatch = True
    layer_modules_strict = False

    dynamic_expert_index = "moe_num_experts"

    pre_lm_head_norm_module = "model.norm"

    module_tree = [
        "model",
        "layers",
        "#",
        {
            "input_layernorm": ("input_layernorm:!",),
            "self_attn": ("q_proj:0", "k_proj:0", "v_proj:0", "o_proj:1"),
            "post_attention_layernorm": ("post_attention_layernorm:!",),
            "mlp:moe": {
                "gate_proj": ("gate_proj",),
                "up_proj": ("up_proj",),
                "down_proj": ("down_proj",),
                "gate": ("gate:!",),
                "shared_experts": {
                    "gate_proj": ("gate_proj:0",),
                    "up_proj": ("up_proj:0",),
                    "down_proj": ("down_proj:1",),
                },
                "experts": {
                    "#": ("gate_proj:0", "upe_proj:0", "down_proj:1"),
                },
            },
        }
    ]

    def monkey_patch(self):
        from dataclasses import dataclass
        from functools import partial
        from typing import Optional

        import torch
        from transformers.cache_utils import Cache, DynamicCache
        from transformers.modeling_flash_attention_utils import FlashAttentionKwargs
        from transformers.modeling_outputs import ModelOutput
        from transformers.processing_utils import Unpack

        @dataclass
        class Erine4_5_MoeModelOutputWithPast(ModelOutput):
            last_hidden_state: Optional[torch.FloatTensor] = None
            past_key_values: Optional[Cache] = None
            hidden_states: Optional[tuple[torch.FloatTensor, ...]] = None
            attentions: Optional[tuple[torch.FloatTensor, ...]] = None
            router_loss: Optional[torch.FloatTensor] = None
            gate_logits: Optional[tuple[torch.FloatTensor, ...]] = None

        def ernie4_5_moe_model_forward(
                self,
                input_ids: Optional[torch.LongTensor] = None,
                attention_mask: Optional[torch.Tensor] = None,
                position_ids: Optional[torch.LongTensor] = None,
                past_key_values: Optional[Cache] = None,
                inputs_embeds: Optional[torch.FloatTensor] = None,
                use_cache: Optional[bool] = None,
                output_attentions: Optional[bool] = None,
                output_hidden_states: Optional[bool] = None,
                cache_position: Optional[torch.LongTensor] = None,
                **flash_attn_kwargs: Unpack[FlashAttentionKwargs],
        ):
            """Forward pass through the ERNIE model."""
            output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
            output_hidden_states = (
                output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
            )

            if (input_ids is None) ^ (inputs_embeds is not None):
                raise ValueError("You must specify exactly one of input_ids or inputs_embeds")

            if self.gradient_checkpointing and self.training:
                if use_cache:
                    use_cache = False

            if use_cache and past_key_values is None:
                past_key_values = DynamicCache()

            if inputs_embeds is None:
                inputs_embeds = self.embed_tokens(input_ids)

            inputs_embeds = inputs_embeds.to(self.embed_tokens.weight.dtype)

            if cache_position is None:
                past_seen_tokens = past_key_values.get_seq_length() if past_key_values is not None else 0
                cache_position = torch.arange(
                    past_seen_tokens, past_seen_tokens + inputs_embeds.shape[1], device=inputs_embeds.device
                )
            if position_ids is None:
                position_ids = cache_position.unsqueeze(0)

            causal_mask = self._update_causal_mask(
                attention_mask, inputs_embeds, cache_position, past_key_values, output_attentions
            )

            hidden_states = inputs_embeds

            # create position embeddings to be shared across the decoder layers
            position_embeddings = self.rotary_emb(hidden_states, position_ids)

            # decoder layers
            all_hidden_states = () if output_hidden_states else None
            all_self_attns = () if output_attentions else None
            all_router_loss = torch.tensor(0.0, device=inputs_embeds.device) if self.config.use_moe else None
            all_gate_logits = ()

            for decoder_layer in self.layers:
                if output_hidden_states:
                    all_hidden_states += (hidden_states,)

                if self.gradient_checkpointing and self.training:
                    layer_outputs = self._gradient_checkpointing_func(
                        partial(decoder_layer.__call__, **flash_attn_kwargs),
                        hidden_states,
                        causal_mask,
                        position_ids,
                        past_key_values,
                        output_attentions,
                        use_cache,
                        cache_position,
                        position_embeddings,
                    )
                else:
                    layer_outputs = decoder_layer(
                        hidden_states=hidden_states,
                        attention_mask=causal_mask,
                        position_ids=position_ids,
                        past_key_value=past_key_values,
                        output_attentions=output_attentions,
                        use_cache=use_cache,
                        cache_position=cache_position,
                        position_embeddings=position_embeddings,
                        **flash_attn_kwargs,
                    )

                hidden_states = layer_outputs[0]

                if output_attentions:
                    all_self_attns += (layer_outputs[1],)

                if self.config.use_moe:
                    layer_outputs, gate_logits = layer_outputs[:-1], layer_outputs[-1]
                    all_gate_logits = all_gate_logits + (gate_logits,)

            hidden_states = self.norm(hidden_states)

            # add hidden states from the last decoder layer
            if output_hidden_states:
                all_hidden_states += (hidden_states,)

            # assert all_router_loss is None, f'moe not support `return-dict`'
            return Erine4_5_MoeModelOutputWithPast(
                last_hidden_state=hidden_states,
                past_key_values=past_key_values,
                hidden_states=all_hidden_states,
                attentions=all_self_attns,
                router_loss=all_router_loss,
                gate_logits=all_gate_logits,
            )

        if not self.load_quantized_model:
            ernie4_5_model = type(self.model.model)
            ernie4_5_model.forward = ernie4_5_moe_model_forward



# SPDX-FileCopyrightText: 2024-2025 NVIDIA CORPORATION
# SPDX-FileCopyrightText: 2025 ModelCloud.ai (qubitium@modelcloud.ai)
# SPDX-License-Identifier: Apache-2.0
# EoRA arXiv https://arxiv.org/abs/2410.21271
# EoRA Official Repo: https://github.com/NVlabs/EoRA
# This file has been modified by ModelCloud.AI team and qubitium@modelcloud.ai for adoption into GPT-QModel

# EoRA
# @article{liu2024eora,
#   title={EoRA: Training-free Compensation for Compressed LLM with Eigenspace Low-Rank Approximation},
#   author={Liu, Shih-Yang and Yang, Huck and Wang, Chien-Yi and Fung, Nai Chit and Yin, Hongxu and Sakr, Charbel and Muralidharan, Saurav and Cheng, Kwang-Ting and Kautz, Jan and Wang, Yu-Chiang Frank and others},
#   journal={arXiv preprint arXiv:2410.21271},
#   year={2024}
# }

from typing import Sequence, Tuple

import torch
from torch import Tensor

from ..utils.logger import setup_logger
from ..utils.rocm import IS_ROCM

log = setup_logger()

def eora_process_input(
        input: Tensor,
        name: str,
        sample_size: int,
        device: torch.device,
) -> Tuple[int, torch.Tensor, float]:
    """Prepare the per-batch covariance contribution required for EoRA.

    The contribution remains on the originating device so multi-GPU execution
    can accumulate locally before a single merge step.
    """

    inp = input[0].to(device=device, dtype=torch.float32)
    if inp.dim() == 2:
        inp = inp.unsqueeze(0)

    batch = inp.shape[0]
    adds = torch.matmul(inp.transpose(1, 2), inp)
    adds_sum = torch.sum(adds, dim=0).detach()

    contribution = adds_sum.to(dtype=torch.float32)
    contribution /= float(sample_size)

    # Adding batch to denominator is only for mathematical stability
    scale = float(sample_size) / (float(sample_size) + float(batch))

    del inp, adds, adds_sum

    return batch, contribution, scale


def merge_eora_segments(segments: Sequence[Tuple[torch.Tensor, float]]) -> torch.Tensor:
    """Combine pre-aggregated EoRA segments using their scale products.

    Each segment entry is a tuple ``(total, scale_product)`` where ``total`` is
    the sequential accumulation result for that segment starting from zero, and
    ``scale_product`` is the product of per-batch scale factors encountered in
    the same segment.  The function mutates the first segment tensor in place
    and returns it as the merged result.
    """
    if not segments:
        raise ValueError("EoRA merge requires at least one segment.")

    result: torch.Tensor | None = None
    for total, scale_product in segments:
        if result is None:
            result = total
        else:
            result.mul_(float(scale_product))
            result.add_(total)

    assert result is not None
    return result

def eora_compute_lora(
        w_wq_delta: Tensor, # need the w (original weight) and wq (quantized qweight) delta in float32
        name: str,
        eigen_scaling_diag_matrix: torch.Tensor,
        rank: int,
        dtype: torch.dtype,
        device: torch.device,
) -> Tuple[Tensor, Tensor]:

    assert w_wq_delta.dtype == torch.float32

    # save this later for SVD
    raw_scaling_diag_matrix = eigen_scaling_diag_matrix.to(device=device, dtype=torch.float64)

    if IS_ROCM:
        # hip cannot resolve linalg ops
        original_backend = torch.backends.cuda.preferred_linalg_library()
        torch.backends.cuda.preferred_linalg_library(backend="magma")

    L, Q = torch.linalg.eigh(raw_scaling_diag_matrix)

    if (L < 0).any():
        ## When expanding the calibration data size for EoRA, I suggest maintaining the balance by allocating 50% to general input (C4) and the remaining 50% to downstream task data.
        log.warn(f"Found negative eigenvalues in `{name}`. Please increase your calibration data set for EoRA.")
        minimum = torch.min(L[L > 0])
        L[L < 0] = minimum

    sqrtEigenvalues = torch.sqrt(L)
    scaling_diag_matrix = Q @ torch.diag(sqrtEigenvalues)

    scaling_matrix_inv = torch.diag(1/sqrtEigenvalues) @ Q.T

    scaling_diag_matrix = scaling_diag_matrix.to(dtype=torch.float32)
    scaling_matrix_inv = scaling_matrix_inv.to(dtype=torch.float32)

    delta_scale = torch.matmul(w_wq_delta, scaling_diag_matrix)

    U, S, V = torch.linalg.svd(delta_scale, full_matrices=False)
    lowrank_r = rank
    truc_s = S[:lowrank_r]
    truc_u = U[:, :lowrank_r]
    truc_v = torch.matmul(V[:lowrank_r, :], scaling_matrix_inv)
    truc_sigma = torch.diag(truc_s)

    sqrtS = torch.sqrt(truc_sigma)
    B = torch.matmul(truc_u, sqrtS).to(dtype=dtype) # default to float16, check if we should save to float32
    A = torch.matmul(sqrtS, truc_v).to(dtype=dtype) # default to float16, check if we should save to float32


    del L, Q, U, S, V,
    del w_wq_delta, raw_scaling_diag_matrix, sqrtEigenvalues, scaling_diag_matrix, scaling_matrix_inv, delta_scale
    del truc_s, truc_u, truc_v, truc_sigma, sqrtS

    # revert linalg backend
    if IS_ROCM:
        torch.backends.cuda.preferred_linalg_library(original_backend)

    return A, B



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

# adapted from @qwopqwop200 's [GPTQ-for-LLaMa](https://github.com/qwopqwop200/GPTQ-for-LLaMa/tree/cuda), which itself is based on [gptq](https://github.com/IST-DASLab/gptq)

import torch
import torch.nn as nn

from ..quantization import QuantizeConfig
from ..utils.logger import setup_logger


log = setup_logger()

HF_OPTIMUM = "hf_optimum"

def quantize(x, scale, zero, maxq, requires_groupwise_processing: bool):
    if maxq < 0:
        return (x > scale / 2).float() * scale + (x < zero / 2).float() * zero
    if requires_groupwise_processing:
        q = torch.clamp(torch.round(x / scale), -maxq, maxq)
        return scale * q
    else:
        q = torch.clamp(torch.round(x / scale) + zero, 0, maxq)
        return scale * (q - zero)


class Quantizer(nn.Module):
    def __init__(self, qcfg: QuantizeConfig, shape=1, name: str=None):
        super(Quantizer, self).__init__()

        self.qcfg = qcfg
        self.register_buffer("maxq", torch.tensor(0))
        self.register_buffer("scale", torch.zeros(shape))
        self.register_buffer("zero", torch.zeros(shape))

        self.name=name

    def requires_groupwise_processing(self) -> bool:
        return False

    # FIXME, optimum shouldn't call this directly, it should call hf_configure
    def configure(
        self,
        perchannel=False,
        grid=100,
        maxshrink=0.8,
        trits=False,
        bits:int=4, # for hf compat
        sym:bool=False, # for hf compat
    ):
        if self.name == HF_OPTIMUM:
            self.qcfg.bits = bits
            self.qcfg.sym = sym

        if self.requires_groupwise_processing():
            self.maxq = torch.tensor(2 ** (self.qcfg.bits - 1) - 1)
        else:
            self.maxq = torch.tensor(2 ** self.qcfg.bits - 1)

        self.perchannel = perchannel
        self.grid = grid
        self.maxshrink = maxshrink
        if trits:
            self.maxq = torch.tensor(-1)

    def find_params(self, x, weight=False):
        dev = x.device
        self.maxq = self.maxq.to(dev)

        shape = x.shape
        if self.perchannel:
            if weight:
                x = x.flatten(1)
            else:
                if len(shape) == 4:
                    x = x.permute([1, 0, 2, 3])
                    x = x.flatten(1)
                if len(shape) == 3:
                    x = x.reshape((-1, shape[-1])).t()
                if len(shape) == 2:
                    x = x.t()
        else:
            x = x.flatten().unsqueeze(0)

        tmp = torch.zeros(x.shape[0], device=dev)
        xmin = torch.minimum(x.min(1)[0], tmp)
        xmax = torch.maximum(x.max(1)[0], tmp)

        if self.qcfg.sym:
            xmax = torch.maximum(torch.abs(xmin), xmax)
            tmp = xmin < 0
            if torch.any(tmp):
                xmin[tmp] = -xmax[tmp]
        tmp = (xmin == 0) & (xmax == 0)
        xmin[tmp] = -1
        xmax[tmp] = +1

        if self.maxq < 0:
            self.scale = xmax
            self.zero = xmin
        else:
            if self.requires_groupwise_processing():
                self.scale = xmax / self.maxq
                self.zero = torch.zeros_like(self.scale)
            else:
                self.scale = (xmax - xmin) / self.maxq
                if self.qcfg.sym:
                    self.zero = torch.full_like(self.scale, (self.maxq + 1) / 2)
                else:
                    self.zero = torch.round(-xmin / self.scale)

        if self.qcfg.mse > 0.0:
            best = torch.full([x.shape[0]], float("inf"), device=dev)
            for i in range(int(self.maxshrink * self.grid)):
                p = 1 - i / self.grid
                xmin1 = p * xmin
                xmax1 = p * xmax
                scale1 = (
                    xmax1 / self.maxq
                    if self.requires_groupwise_processing()
                    else (xmax1 - xmin1) / self.maxq
                )
                zero1 = torch.round(-xmin1 / scale1) if not self.qcfg.sym else self.zero
                q = quantize(x, scale1.unsqueeze(1), zero1.unsqueeze(1), self.maxq, self.requires_groupwise_processing())
                q -= x
                q.abs_()
                q.pow_(self.qcfg.mse)
                err = torch.sum(q, 1)
                tmp = err < best
                if torch.any(tmp):
                    best[tmp] = err[tmp]
                    self.scale[tmp] = scale1[tmp]
                    self.zero[tmp] = zero1[tmp]
        if not self.perchannel:
            if weight:
                tmp = shape[0]
            else:
                tmp = shape[1] if len(shape) != 3 else shape[2]
            self.scale = self.scale.repeat(tmp)
            self.zero = self.zero.repeat(tmp)

        if weight:
            shape = [-1] + [1] * (len(shape) - 1)
            self.scale = self.scale.reshape(shape)
            self.zero = self.zero.reshape(shape)
            return
        if len(shape) == 4:
            self.scale = self.scale.reshape((1, -1, 1, 1))
            self.zero = self.zero.reshape((1, -1, 1, 1))
        if len(shape) == 3:
            self.scale = self.scale.reshape((1, 1, -1))
            self.zero = self.zero.reshape((1, 1, -1))
        if len(shape) == 2:
            self.scale = self.scale.unsqueeze(0)
            self.zero = self.zero.unsqueeze(0)

    def quantize(self, x):
        return quantize(x, self.scale, self.zero, self.maxq, self.requires_groupwise_processing())

    # def enabled(self):
    #     return self.maxq > 0

    # def ready(self):
    # return torch.all(self.scale != 0)

class QQQQuantizer(Quantizer):
    def requires_groupwise_processing(self) -> bool:
        return self.qcfg.group_size == -1 and self.qcfg.sym

__all__ = ["Quantizer"]



# SPDX-FileCopyrightText: 2024-2025 ModelCloud.ai
# SPDX-FileCopyrightText: 2024-2025 qubitium@modelcloud.ai
# SPDX-License-Identifier: Apache-2.0
# Contact: qubitium@modelcloud.ai, x.com/qubitium

from contextlib import nullcontext
from typing import Optional, Tuple

import torch

from ...adapter.adapter import Adapter, Lora
from ...models._const import DEVICE, PLATFORM
from ...nn_modules.qlinear import AWQuantLinear
from ...quantization import FORMAT, METHOD
from ...utils import has_gil_disabled
from ...utils.backend import BACKEND
from ...utils.torch import HAS_XPU


class AwqGemmTritonFn(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx,
        x,
        qweight,
        qzeros,
        scales,
        w_bit=4,
        group_size=128,
        bias=None,
        out_features=0,
        prefer_backend=None,
    ):
        from ...quantization.awq.modules.triton.gemm import awq_dequantize_triton, awq_gemm_triton

        ctx.save_for_backward(x, qweight, qzeros, scales, bias)
        ctx.out_features = out_features

        out_shape = x.shape[:-1] + (out_features,)
        x = x.to(torch.float16)
        if x.shape[0] == 0:
            return torch.zeros(out_shape, dtype=x.dtype, device=x.device)

        # Above compute density threshold it is faster to just dequantize the whole thing and do simple matmul
        FULL_DEQUANT_MATMUL_THRESHOLD = x.shape[0] * x.shape[1] > 128
        if FULL_DEQUANT_MATMUL_THRESHOLD:
            out = awq_dequantize_triton(qweight, scales, qzeros)
            out = torch.matmul(x, out.to(x.dtype))
        else:
            out = awq_gemm_triton(
                x.reshape(-1, x.shape[-1]), qweight, scales, qzeros, split_k_iters=8,
            )

        out = out + bias if bias is not None else out
        out = out.reshape(out_shape)
        if len(out.shape) == 2:
            out = out.unsqueeze(0)
        return out

    @staticmethod
    def backward(ctx, grad_output):
        from ...quantization.awq.modules.triton.gemm import awq_dequantize_triton


        input, qweight, qzeros, scales, bias = ctx.saved_tensors

        weights = awq_dequantize_triton(qweight, scales, qzeros).to(grad_output.dtype)

        grad_input = None
        if ctx.needs_input_grad[0]:
            batch_size = grad_output.shape[0]
            grad_input = grad_output.bmm(weights.transpose(0, 1).unsqueeze(0).repeat(batch_size, 1, 1))

        return grad_input, None, None, None, None, None, None, None, None


class AwqGEMMTritonQuantLinear(AWQuantLinear):
    SUPPORTS_BACKENDS = [BACKEND.GEMM_TRITON]
    SUPPORTS_METHODS = [METHOD.AWQ]
    SUPPORTS_FORMATS = {FORMAT.GEMM: 50}
    SUPPORTS_BITS = [4]
    SUPPORTS_GROUP_SIZE = [-1, 16, 32, 64, 128]
    SUPPORTS_DESC_ACT = [True, False]
    SUPPORTS_SYM = [True, False]
    SUPPORTS_SHARDS = True
    SUPPORTS_TRAINING = True
    SUPPORTS_AUTO_PADDING = False
    SUPPORTS_IN_FEATURES_DIVISIBLE_BY = [1]
    SUPPORTS_OUT_FEATURES_DIVISIBLE_BY = [1]

    # TODO: ROCM also has Triton support. Need to validate ROCM for triton
    SUPPORTS_DEVICES = [DEVICE.CUDA]
    SUPPORTS_PLATFORM = [PLATFORM.LINUX, PLATFORM.WIN32]
    SUPPORTS_PACK_DTYPES = [torch.int32]
    SUPPORTS_ADAPTERS = [Lora]

    SUPPORTS_DTYPES = [torch.float16]

    REQUIRES_FORMAT_V2 = False

    QUANT_TYPE = "awq_gemm_triton"

    @classmethod
    def validate_once(cls) -> Tuple[bool, Optional[Exception]]:
        from packaging import version
        from triton import __version__ as triton_version
        triton_v = version.parse(triton_version)

        if triton_v < version.parse("2.0.0"):
            raise ImportError(f"triton version must be >= 2.0.0: actual = {triton_version}")

        # GIL=0 is tested with Triton 3.4.0 and it works
        if has_gil_disabled() and triton_v < version.parse("3.4.0"):
            raise Exception("GIL is disabled and not compatible with current Triton. Please upgrade to Triton >= 3.4.0")

        return True, None

    def __init__(
        self,
        bits: int,
        group_size: int,
        sym: bool,
        desc_act: bool,
        in_features: int,
        out_features: int,
        bias: bool = False,
        pack_dtype: torch.dtype = torch.int32,
        adapter: Adapter = None,
        register_buffers: bool = False,
        **kwargs,
    ):
        super().__init__(
            bits=bits,
            group_size=group_size,
            sym=sym,
            desc_act=desc_act,
            in_features=in_features,
            out_features=out_features,
            bias=bias,
            pack_dtype=pack_dtype,
            backend=kwargs.pop("backend", BACKEND.TRITON),
            adapter=adapter,
            register_buffers=register_buffers,
            **kwargs)

    def post_init(self):
        if self.scales is not None:
            self.scales = self.scales.to(dtype=torch.float16)
        super().post_init()

    def forward(self, x: torch.Tensor):
        out_shape = x.shape[:-1] + (self.out_features,)

        input_dtype = x.dtype
        if input_dtype != torch.float16:
            x = x.half()

        with torch.xpu.device(self.qweight.device) if HAS_XPU else torch.cuda.device(self.qweight.device):
            with nullcontext() if self.training else torch.inference_mode():
                out = AwqGemmTritonFn.apply(
                    x,
                    self.qweight,
                    self.qzeros,
                    self.scales,
                    self.bits,
                    self.group_size,
                    self.bias,
                    self.out_features,
                    "triton",
                )

        if input_dtype != torch.float16:
            out = out.to(dtype=input_dtype)

        if self.adapter:
            out = self.adapter.apply(x=x, out=out)

        return out.reshape(out_shape)


__all__ = [
    "AwqGemmTritonFn",
    "AwqGEMMTritonQuantLinear",
]
